from dagster import Array, Field, Noneable, Selector, Shape
from dagster.config.field import resolve_to_config_type
from dagster.core.meta.config_types import (
    ConfigTypeKind,
    ConfigTypeMeta,
    NonGenericTypeRefMeta,
    meta_from_config_type,
)
from dagster.core.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple


def meta_from_dagster_type(dagster_type):
    return meta_from_config_type(resolve_to_config_type(dagster_type))


def test_basic_int_meta():
    int_meta = meta_from_dagster_type(int)
    assert int_meta.given_name == 'Int'
    assert int_meta.key == 'Int'
    assert int_meta.kind == ConfigTypeKind.SCALAR
    assert int_meta.inner_type_refs == []
    assert int_meta.enum_values is None
    assert int_meta.fields is None


def test_basic_dict():
    dict_meta = meta_from_dagster_type({'foo': int})
    assert dict_meta.key.startswith('Shape.')
    assert dict_meta.given_name is None
    assert dict_meta.inner_type_refs
    assert len(dict_meta.inner_type_refs) == 1
    assert dict_meta.inner_type_refs[0].key == 'Int'
    assert dict_meta.enum_values is None

    assert dict_meta.fields and len(dict_meta.fields) == 1

    field = dict_meta.fields[0]
    assert field.name == 'foo'


def test_field_things():
    dict_meta = meta_from_dagster_type(
        {
            'req': int,
            'opt': Field(int, is_required=False),
            'opt_with_default': Field(int, is_required=False, default_value=2),
            'req_with_desc': Field(int, description='A desc'),
        }
    )

    assert dict_meta.fields and len(dict_meta.fields) == 4

    field_meta_dict = {field_meta.name: field_meta for field_meta in dict_meta.fields}

    assert field_meta_dict['req'].is_required is True
    assert field_meta_dict['req'].description is None
    assert field_meta_dict['opt'].is_required is False
    assert field_meta_dict['opt_with_default'].is_required is False
    assert field_meta_dict['opt_with_default'].default_provided is True
    assert field_meta_dict['opt_with_default'].default_value_as_str == '2'

    assert field_meta_dict['req_with_desc'].is_required is True
    assert field_meta_dict['req_with_desc'].description == 'A desc'


def test_basic_list():
    list_meta = meta_from_dagster_type(Array(int))
    assert list_meta.key.startswith('Array')
    assert list_meta.inner_type_refs
    assert len(list_meta.inner_type_refs) == 1
    assert list_meta.inner_type_refs[0].key == 'Int'
    assert list_meta.enum_values is None


def test_basic_optional():
    optional_meta = meta_from_dagster_type(Noneable(int))
    assert optional_meta.key.startswith('Noneable')
    assert optional_meta.inner_type_refs
    assert len(optional_meta.inner_type_refs) == 1
    assert optional_meta.inner_type_refs[0].key == 'Int'
    # https://github.com/dagster-io/dagster/issues/1933
    # TODO reconcile names
    assert optional_meta.kind == ConfigTypeKind.NONEABLE
    assert optional_meta.enum_values is None


def test_basic_list_list():
    list_meta = meta_from_dagster_type([[int]])
    assert list_meta.key.startswith('Array')
    assert list_meta.inner_type_refs
    assert len(list_meta.inner_type_refs) == 2
    refs = {ref.key: ref for ref in list_meta.inner_type_refs}
    assert (
        len(refs['Array.Int'].inner_type_refs) == 1
        and isinstance(refs['Array.Int'], ConfigTypeMeta)
        and refs['Array.Int'].inner_type_refs[0].key == 'Int'
    )
    assert refs['Int'].key == 'Int'
    assert list_meta.enum_values is None

    assert (
        len(list_meta.type_param_refs) == 1
        and list_meta.type_param_refs[0].kind == ConfigTypeKind.ARRAY
    )


def test_list_of_dict():
    inner_dict_dagster_type = Shape({'foo': Field(str)})
    list_of_dict_meta = meta_from_dagster_type([inner_dict_dagster_type])

    assert list_of_dict_meta.key.startswith('Array')
    assert list_of_dict_meta.inner_type_refs
    assert len(list_of_dict_meta.inner_type_refs) == 1
    # Both Shape[...] and str are NonGenericTypeRefMetas in this schema
    dict_ref = list_of_dict_meta.type_param_refs[0]
    assert isinstance(dict_ref, NonGenericTypeRefMeta)
    assert dict_ref.key.startswith('Shape')

    assert (
        len(list_of_dict_meta.type_param_refs) == 1
        and list_of_dict_meta.type_param_refs[0].key
        == resolve_to_config_type(inner_dict_dagster_type).key
    )


def test_selector_of_things():
    selector_meta = meta_from_dagster_type(Selector({'bar': Field(int)}))
    assert selector_meta.key.startswith('Selector')
    assert selector_meta.kind == ConfigTypeKind.SELECTOR
    assert selector_meta.fields and len(selector_meta.fields) == 1
    field_meta = selector_meta.fields[0]
    assert field_meta.name == 'bar'
    assert field_meta.type_ref.key == 'Int'


def test_kitchen_sink():
    kitchen_sink = resolve_to_config_type(
        [
            {
                'opt_list_of_int': Field(int, is_required=False),
                'nested_dict': {
                    'list_list': [[int]],
                    'nested_selector': Field(
                        Selector({'some_field': int, 'more_list': Noneable([bool])})
                    ),
                },
            }
        ]
    )

    kitchen_sink_meta = meta_from_dagster_type(kitchen_sink)

    rehydrated_meta = deserialize_json_to_dagster_namedtuple(
        serialize_dagster_namedtuple(kitchen_sink_meta)
    )
    assert kitchen_sink_meta == rehydrated_meta


def test_kitchen_sink_break_out():
    nested_dict_cls = resolve_to_config_type(
        {
            'list_list': [[int]],
            'nested_selector': Selector({'some_field': int, 'list': Noneable([bool])}),
        }
    )
    dict_within_list_cls = resolve_to_config_type(
        {'opt_list_of_int': Field([int], is_required=False), 'nested_dict': Field(nested_dict_cls)}
    )
    kitchen_sink = Array(dict_within_list_cls)

    dict_within_list_key = dict_within_list_cls.key
    kitchen_sink_meta = meta_from_dagster_type(kitchen_sink)

    assert len(kitchen_sink_meta.type_param_refs) == 1
    assert kitchen_sink_meta.type_param_refs[0].key == dict_within_list_key
    assert len(kitchen_sink_meta.inner_type_refs) == 1
    assert kitchen_sink_meta.inner_type_refs[0].key == dict_within_list_key
    dict_within_list_meta = meta_from_dagster_type(dict_within_list_cls)
    assert dict_within_list_meta.type_param_refs is None
    # List[int], Int, Shape.XXX
    assert len(dict_within_list_meta.inner_type_refs) == 3
    assert sorted([type_ref.key for type_ref in dict_within_list_meta.inner_type_refs]) == sorted(
        [nested_dict_cls.key, 'Int', 'Array.Int']
    )



import os

from dagster import check
from dagster.config.config_type import ConfigStringInstance, ScalarUnion
from dagster.config.field_utils import Selector


class StringSourceType(ScalarUnion):
    def __init__(self):
        super(StringSourceType, self).__init__(
            scalar_type=ConfigStringInstance,
            non_scalar_type=Selector({'env': str}),
            _key='StringSourceType',
        )

    def post_process(self, value):
        if not isinstance(value, dict):
            return value

        key, cfg = list(value.items())[0]
        if key == 'env':
            value = os.getenv(cfg)
            check.invariant(
                value is not None, 'Environment variable "{var}" is not set.'.format(var=cfg)
            )
            return value
        else:
            check.failed('Invalid source selector key')


StringSource = StringSourceType()



import io
import os
import shutil
import warnings
from collections import OrderedDict

import six

from dagster import check, utils
from dagster.core.definitions import RepositoryDefinition
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.scheduler import Schedule
from dagster.core.serdes import (
    ConfigurableClass,
    deserialize_json_to_dagster_namedtuple,
    serialize_dagster_namedtuple,
)

from .base import ScheduleStorage


class FilesystemScheduleStorage(ScheduleStorage, ConfigurableClass):
    def __init__(self, base_dir, inst_data=None):
        self._base_dir = check.str_param(base_dir, 'base_dir')
        self._inst_data = inst_data
        self._schedules = OrderedDict()
        self._load_schedules()

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {'base_dir': str}

    @staticmethod
    def from_config_value(inst_data, config_value):
        return FilesystemScheduleStorage.from_local(inst_data=inst_data, **config_value)

    @staticmethod
    def from_local(base_dir, inst_data=None):
        return FilesystemScheduleStorage(base_dir, inst_data)

    def all_schedules(self, repository=None):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        # TODO: Check for repository=None

        if repository.name not in self._schedules:
            return []

        return [s for s in self._schedules[repository.name].values()]

    def get_schedule_by_name(self, repository, schedule_name):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.str_param(schedule_name, 'schedule_name')

        if repository.name not in self._schedules:
            return None

        return self._schedules[repository.name].get(schedule_name)

    def add_schedule(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        if not repository.name in self._schedules:
            self._schedules[repository.name] = OrderedDict()

        self._schedules[repository.name][schedule.name] = schedule
        self._write_schedule_to_file(repository, schedule)

    def update_schedule(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        if repository.name not in self._schedules:
            raise DagsterInvariantViolationError(
                'Repository {repository_name} is not present in storage'.format(
                    repository_name=repository.name
                )
            )

        if schedule.name not in self._schedules[repository.name]:
            raise DagsterInvariantViolationError(
                'Schedule {name} is not present in storage'.format(name=schedule.name)
            )

        self.add_schedule(repository, schedule)

    def delete_schedule(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        if repository.name not in self._schedules:
            raise DagsterInvariantViolationError(
                'Repository {repository_name} is not present in storage'.format(
                    repository_name=repository.name
                )
            )

        if schedule.name not in self._schedules[repository.name]:
            raise DagsterInvariantViolationError(
                'Schedule {name} is not present in storage'.format(name=schedule.name)
            )

        self._schedules[repository.name].pop(schedule.name)
        self._delete_schedule_file(repository, schedule)

    def wipe(self):
        self._schedules = OrderedDict()
        for repository_name in self._schedules.keys():
            shutil.rmtree(os.path.join(self._base_dir, repository_name))

    def _write_schedule_to_file(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        repository_folder = os.path.join(self._base_dir, repository.name)
        utils.mkdir_p(repository_folder)

        metadata_file = os.path.join(repository_folder, '{}.json'.format(schedule.name),)

        with io.open(metadata_file, 'w+', encoding='utf-8') as f:
            f.write(six.text_type(serialize_dagster_namedtuple(schedule)))

        return metadata_file

    def _delete_schedule_file(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        metadata_file = os.path.join(
            self._base_dir, repository.name, '{}.json'.format(schedule.name),
        )

        os.remove(metadata_file)

    def _load_schedules(self):
        schedules_dir = os.path.join(self._base_dir)
        utils.mkdir_p(schedules_dir)

        for repository_name in os.listdir(schedules_dir):
            if not os.path.isdir(os.path.join(schedules_dir, repository_name)):
                continue

            self._schedules[repository_name] = {}
            for file in os.listdir(os.path.join(schedules_dir, repository_name)):
                if not file.endswith('.json'):
                    continue
                file_path = os.path.join(schedules_dir, repository_name, file)
                with open(file_path) as data:
                    try:
                        schedule = deserialize_json_to_dagster_namedtuple(data.read())
                        self._schedules[repository_name][schedule.name] = schedule

                    except Exception as ex:  # pylint: disable=broad-except
                        warnings.warn(
                            'Could not parse dagster schedule from {file_name} in {dir_name}. '
                            '{ex}: {msg}'.format(
                                file_name=file,
                                dir_name=self._base_dir,
                                ex=type(ex).__name__,
                                msg=ex,
                            )
                        )
                        continue



from dagster import check

from .config_type import ConfigType, ConfigTypeKind


def iterate_config_types(config_type):

    # TODO: Investigate during config refactor. Very sketchy -- schrockn 12/05/19
    # Looping over this should be done at callsites
    if isinstance(config_type, list):
        check.list_param(config_type, 'config_type', of_type=ConfigType)
        for config_type_item in config_type:
            for inner_type in iterate_config_types(config_type_item):
                yield inner_type

    check.inst_param(config_type, 'config_type', ConfigType)
    if config_type.kind == ConfigTypeKind.ARRAY or config_type.kind == ConfigTypeKind.NONEABLE:
        for inner_type in iterate_config_types(config_type.inner_type):
            yield inner_type

    if ConfigTypeKind.has_fields(config_type.kind):
        for field_type in config_type.fields.values():
            for inner_type in iterate_config_types(field_type.config_type):
                yield inner_type

    if config_type.kind == ConfigTypeKind.SCALAR_UNION:
        yield config_type.scalar_type
        yield config_type.non_scalar_type

    yield config_type



from collections import defaultdict

import gevent.lock

from dagster import check
from dagster.core.events.log import EventRecord

from .base import EventLogSequence, EventLogStorage


class InMemoryEventLogStorage(EventLogStorage):
    def __init__(self):
        self._logs = defaultdict(EventLogSequence)
        self._lock = defaultdict(gevent.lock.Semaphore)
        self._handlers = defaultdict(set)

    def get_logs_for_run(self, run_id, cursor=-1):
        check.str_param(run_id, 'run_id')
        check.int_param(cursor, 'cursor')
        check.invariant(
            cursor >= -1,
            'Don\'t know what to do with negative cursor {cursor}'.format(cursor=cursor),
        )

        cursor = cursor + 1
        with self._lock[run_id]:
            return self._logs[run_id][cursor:]

    def store_event(self, event):
        check.inst_param(event, 'event', EventRecord)
        run_id = event.run_id
        with self._lock[run_id]:
            self._logs[run_id] = self._logs[run_id].append(event)
            for handler in self._handlers[run_id]:
                handler(event)

    def delete_events(self, run_id):
        with self._lock[run_id]:
            del self._logs[run_id]
        del self._lock[run_id]

    def wipe(self):
        self._logs = defaultdict(EventLogSequence)
        self._lock = defaultdict(gevent.lock.Semaphore)

    def watch(self, run_id, _start_cursor, callback):
        with self._lock[run_id]:
            self._handlers[run_id].add(callback)

    def end_watch(self, run_id, handler):
        with self._lock[run_id]:
            if handler in self._handlers[run_id]:
                self._handlers[run_id].remove(handler)

    @property
    def is_persistent(self):
        return False
